use anyhow::{bail, Result};
use serde::{Deserialize, Serialize};

use super::super::errors::*;
use super::super::sps_datatypes::{BitPosition, DataEvaluation, UnparsedSPSDataType};

#[derive(Debug, Serialize, Deserialize, Copy, Clone, PartialEq, Eq, Hash)]
pub struct DateTimeType {
    length: u32,
    position: BitPosition,
}
impl DateTimeType {
    const LEN: usize = 8;
    pub fn new(byte: u32, bit: Option<u32>) -> Self {
        DateTimeType {
            length: Self::LEN as u32,
            position: BitPosition { byte, bit },
        }
    }

    fn assert_range_inclusive(&self, input: u8, min: u8, max: u8, field: &str) -> Result<u8> {
        if input < min {
            bail!(
                "Value '{}' is lower than the minimum '{}' allowed for {}.",
                input,
                min,
                field
            )
        }
        if input > max {
            bail!(
                "Value '{}' is higher than the maximum '{}' allowed for {}.",
                input,
                max,
                field
            )
        }
        Ok(input)
    }

    fn parse(&self, data: &[u8]) -> Result<String> {
        let year: Result<u32> = match data[0] {
            e if e < 90 => Ok(e as u32 + 2000),
            e if e < 100 => Ok(e as u32 + 1900),
            e => bail!(
                "Value '{}' is higher than the maximum '99' of S7 date and time representation.",
                e
            ),
        };
        let month = self.assert_range_inclusive(data[1], 1, 12, "month")?;
        let day = self.assert_range_inclusive(data[2], 1, 31, "day of month")?;
        let hour = self.assert_range_inclusive(data[3], 0, 23, "hour")?;
        let minute = self.assert_range_inclusive(data[4], 0, 59, "minute")?;
        let second = self.assert_range_inclusive(data[5], 0, 59, "second")?;
        let _sec = self.assert_range_inclusive(data[6], 0, 99, "first two millisecond digits")?;
        let _msec = self.assert_range_inclusive(data[7] >> 4, 0, 9, "third millisecond digit")?;
        let _day_of_week =
            self.assert_range_inclusive(data[7] & 0b00001111, 1, 7, "day of week")?;

        Ok(format!(
            "{}-{:02}-{:02} {:02}:{:02}:{:02}",
            year?, month, day, hour, minute, second
        ))
    }
}
impl DataEvaluation for DateTimeType {
    fn into_unparsed(&self) -> UnparsedSPSDataType {
        UnparsedSPSDataType {
            data_type: self.into_string(),
            data_byte: self.position.byte,
            data_bit: self.position.bit,
            data_length: None,
        }
    }
    fn into_string(&self) -> String {
        "datetime".to_string()
    }
    fn get_end_byte(&self) -> u32 {
        self.position.byte + self.length
    }
    fn get_byte_positon(&self) -> u32 {
        self.position.byte
    }
    fn get_length(&self) -> u32 {
        self.length
    }
    fn parse_serial(&self, data: &[&str]) -> Result<String> {
        Ok(data
            .get((self.position.byte) as usize)
            .ok_or_else(|| serial_error(self.into_string(), self.position.byte))?
            .to_string())
    }
    fn parse_s7(&self, data: &[u8]) -> Result<String> {
        let bytes = data
            .get((self.position.byte) as usize..(self.position.byte + self.length) as usize)
            .ok_or_else(|| s7_read_error(self.into_string(), self.position.byte))?;
        self.parse(&bytes.to_vec())
    }

    fn sql_equivalent(&self) -> &str {
        r"DATETIME DEFAULT NULL"
    }
}

#[test]
fn test() {
    const DATEPOS: u32 = 5;
    const LEN: u32 = 8;
    const VAL: &str = "1993-12-25 08:12:34";

    let test_item = DateTimeType::new(DATEPOS, None);
    let raw_data: [u8; 50] = [255; 50];
    let mut test_vec: Vec<u8> = raw_data.to_vec();
    test_vec.splice(
        DATEPOS as usize..DATEPOS as usize + LEN as usize,
        [93, 12, 25, 8, 12, 34, 56, 7 << 4 | 7],
    );

    let result = match test_item.parse_s7(&test_vec) {
        Ok(res) => res,
        Err(_) => "Error".to_string(),
    };

    assert_eq!(result, VAL.to_string())
}
